require "socket.core"

--Helpers
local function hexstr(p)
	return ("%02X"):rep(#p):format(p:byte(1,#p))
end
local function leUInt(n)
	return string.char(n%256)..string.char(math.floor(n/256)%256)..string.char(math.floor(n/65536)%256)..string.char(math.floor(n/16777216)%256)
end
local function rxUInt(n)
	local a,b,c,d=n:byte(1,4)
	return a+(256*b)+(65536*c)+(16777216*d)
end
local function rxString(s,o)
	o=o or 1
	local os=o
	while s:sub(o,o):byte(1)~=0 do
		o=o+1
	end
	local r=s:sub(os,o-1)
	return r,s:sub(o+1)
end

--Probe outbound IP
local probe=socket.udp()
probe:setpeername("8.8.8.8",1)
local myip=probe:getsockname()
probe=nil

--Advertisements
local adverts=socket.udp()
adverts:setoption('reuseport',true)
adverts:setoption('reuseaddr',true)
adverts:setsockname(myip,15000)
adverts:settimeout(0)

local function advertsPoll()
	local adv=adverts:receive()
	if adv and #adv>=16 and adv:sub(1,8)=="Gideros0" then
		local host=adv:byte(9).."."..adv:byte(10).."."..adv:byte(11).."."..adv:byte(12)
		local port=adv:byte(13)*256+adv:byte(14)
		local flags=adv:byte(15)*256+adv:byte(16)
		local name=adv:sub(17)
		if host=="0.0.0.0" then host=myip end
		return host,port,flags,name
	end
end

-- TCP
local GSocket={}
GSocket.__index=GSocket

function GSocket:init(host,port,client)
	self.s=socket.tcp()
	ide:Print("Connect:",self.s:connect(host,port))
	self.s:settimeout(0.01)
	self.client=client
	self._id=1
	self._rxdata=""
	self.breakpoints={}
end

function GSocket:disconnect()
	self.s:close()
	self.s=nil
end

function GSocket:sendData(data,type)
	data=data or ""
	local id=self._id
	self._id=self._id+1
	local header_sz=#data+12
	local hdr=leUInt(header_sz)..leUInt(id)..leUInt(type)
	self.s:send(hdr..data)
	return id
end

function GSocket:sendCommand(data)
	return self:sendData(data,0)
end

function GSocket:sendAck(id)
	return self:sendData(leUInt(id),1)
end

function GSocket:createFolder(foldername)
	return self:sendCommand(string.char(0) --Create Folder CMD
			..foldername..string.char(0))
end

function GSocket:sendFile(filename,filedata)
	return self:sendCommand(string.char(1) --SendFile CMD
			..filename..string.char(0)..filedata)
end

function GSocket:play(luafiles)
	local cmd=string.char(2) --PLAY CMD
	for _,f in ipairs(luafiles) do
		cmd=cmd..f..string.char(0)
	end
	return self:sendCommand(cmd)
end

function GSocket:stop()
	local cmd=string.char(3) --STOP CMD
	return self:sendCommand(cmd)
end

function GSocket:deleteFiles()
	local cmd=string.char(5) --DELETE FILES CMD
	return self:sendCommand(cmd)
end

function GSocket:getFiles()
	local cmd=string.char(7) --GET FILES CMD
	return self:sendCommand(cmd)
end

function GSocket:setProjectName(projectName)
	local cmd=string.char(8) --PROJECT NAME CMD
		..projectName..string.char(0)
	return self:sendCommand(cmd)
end

function GSocket:deleteFile(filename)
	return self:sendCommand(string.char(9) --Delete file CMD
			..filename..string.char(0))
end

function GSocket:setProjectProperties(properties)
	properties.imageScales=properties.imageScales or {}
	local cmd=string.char(11) --PROJECT PROPERTIES CMD
		..leUInt(properties.scaleMode or 0)
		..leUInt(properties.logicalWidth or 240)
		..leUInt(properties.logicalHeight or 320)
		..leUInt(#properties.imageScales)
	for _,p in ipairs(properties.imageScales) do
		cmd=cmd..p[1]..string.char(0)..leFloat(p[2])
	end
	cmd=cmd..leUInt(properties.orientation or 0)
		..leUInt(properties.fps or 60)
		..leUInt(properties.retinaDisplay or 0)
		..leUInt(properties.autorotation or 0)
		..leUInt(properties.mouseToTouch or 1)
		..leUInt(properties.touchToMouse or 1)
		..leUInt(properties.mouseTouchOrder or 0)
		..leUInt(properties.windowWidth or 0)
		..leUInt(properties.windowHeight or 0)
	return self:sendCommand(cmd)
end

--DEBUG
function GSocket:setBreakpoints(breakpoints)
	self.breakpoints=breakpoints or {}
end
	
function GSocket:lookupSymbol(symbol)
	local cmd=string.char(20) --LOOKUP SYMBOL CMD
		..symbol..string.char(0)
	return self:sendCommand(cmd)
end

function GSocket:startDebugging()
	local cmd=string.char(21) --START DEBUG CMD
		..string.char(0x84)
		..leUInt(#self.breakpoints)
	for _,p in ipairs(self.breakpoints) do
		cmd=cmd..p.line..string.char(0)..p.source..string.char(0)
	end
	return self:sendCommand(cmd)
end

function GSocket:resumeDebug(mode)
	local cmd=string.char(22) --RESUME DEBUG CMD
		..string.char(mode)
		..leUInt(#self.breakpoints)
	for _,p in ipairs(self.breakpoints) do
		cmd=cmd..p.line..string.char(0)..p.source..string.char(0)
	end
	return self:sendCommand(cmd)
end

function GSocket:resume()
	self:resumeDebug(0x84)
end

function GSocket:stepOver()
	self:resumeDebug(0x44)
end

function GSocket:stepInto()
	self:resumeDebug(0x04)
end

function GSocket:stepReturn()
	self:resumeDebug(0x46)
end

function GSocket:stopDebugging()
	self:resumeDebug(0x00)
end

-- PROFLING
function GSocket:startProfiling()
	local cmd=string.char(30) --START PROFILING CMD
	return self:sendCommand(cmd)
end

-- RX
function GSocket:deserialize(data)
	local ltype=data:byte(1)
	if ltype==255 or ltype==0 then return nil,data:sub(2) end --none,nil
	if ltype==1 then return data:byte(2)~=0,data:sub(3) end --boolean
	if ltype==6 then --table
		local ref,tbl=nil,{}
		ref,data=rxString(data,2)
		while true do
			local k,v
			k,data=self:deserialize(data)
			if k==nil then break end
			v,data=self:deserialize(data)
			tbl[k]=v
		end
		return tbl,data
	end
	if ltype==3 then --number
		local n
		n,data=rxString(data,2)
		return tonumber(n),data
	end
	if ltype==5 then --string
		local n
		n,data=rxString(data,2)
		return tostring(n),data
	end
	--Other, generic
	local n
	n,data=rxString(data,2)
	return n,data
end

function GSocket:handleCommand(cmd)
	if #cmd==0 then return end
	local code=cmd:byte(1)
	if code==4 then -- PRINT
		self.client:print(rxString(cmd,2))
	elseif code==6 then -- GOT FILE LIST
		cmd=cmd:sub(2)
		local files={}
		while #cmd>0 do
			local fn
			fn,cmd=rxString(cmd)
			if fn:sub(1,1)=="F" then
				local age=rxUInt(cmd:sub(1,4))
				local md5=cmd:sub(5,20)
				cmd=cmd:sub(21)
				table.insert(files,{file=fn:sub(2), md5=md5, age=age})
			elseif fn:sub(1,1)=="D" then
				table.insert(files,{folder=fn:sub(2)})
			end
		end
		self.client:fileList(files)
	elseif code==23 then -- BREAKED
		local line
		line,cmd=rxString(cmd,2)
		local source=rxString(cmd)
		self.client:breaked(line,source)
	elseif code==24 then -- LOOKED UP SYMBOL
		self.client:symbol(self:deserialize(cmd:sub(2)))
	elseif code==31 then -- PROFILING REPORT
		self.client:profilingReport(self:deserialize(cmd:sub(2)))
	end
end
	
function GSocket:poll()
	local peer=self.s and self.s:getpeername()
	if self.peer and not peer then
		self.client:disconnected()
	elseif peer and not self.peer then
		self.client:connected()
	end
	self.peer=peer
	if self.peer then
		local cdata,_,pdata=self.s:receive("*a")
		if cdata or pdata then
			local data=cdata or pdata
			if #data>0 then
				--ide:Print("TRX:"..tostring(cdata and #cdata).."/"..tostring(pdata and #pdata))
				self._rxdata=self._rxdata..data
				--print("GB:",#self._rxdata)
			end
		end
		while #self._rxdata>=12 do
			local hsize=rxUInt(self._rxdata:sub(1,4))
			--print("GRR:",hsize,#self._rxdata)
			if #self._rxdata>=hsize then
				--local hid=rxUInt(self._rxdata:sub(5,8))
				local htype=rxUInt(self._rxdata:sub(9,12))
				local hdata=self._rxdata:sub(13,hsize)
				self._rxdata=self._rxdata:sub(hsize+1)
				--print("GRX:",hid,htype,hexstr(hdata))
				if htype==1 then
					self.client:ack(rxUInt(hdata))
				elseif htype==0 then
					self:handleCommand(hdata)
				end
			end
		end
	end
end
	
return {
	poll=advertsPoll,
	connect=function(...)
		local con={}
		setmetatable(con,GSocket)
		con:init(...)
		return con
	end,
}